#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <omp.h>

#define N 40000000

int main (void) {
  int i;
  double timer;

  double dx;
  double * v;
  double * u;

  v = malloc(N * sizeof(double));
  u = malloc(N * sizeof(double));

  if (v == NULL || u == NULL) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }

  dx = 1.0/N;

#pragma omp parallel for
  for (i = 0; i < N; i++) {
    v[i] = sin(i*dx);
    u[i] = cos(i*dx);
  }

  /* black magic */
#define DO_PRAGMA(x) _Pragma (#x)

  /* abandon all hope ye who enter here */
#define TIME_IMPLEMENTATION(sched, chunk)	\
  do {						\
    double s = 0.0;				\
						\
    timer = omp_get_wtime();			\
						\
    DO_PRAGMA(omp parallel for reduction(+:s)	\
	      schedule(sched, chunk))		\
      for (i = 0; i < N; i++)			\
	s += v[i]*u[i];				\
						\
    s /= N;					\
						\
    timer = omp_get_wtime() - timer;		\
						\
    printf("Schedule: %s\n"			\
	   "Chunksize: %d\n"			\
	   "Time: %f\n"				\
	   "s: %f\n\n",				\
           #sched,				\
	   chunk,				\
	   timer,				\
	   s);					\
    fflush(stdout);				\
  } while (0)

  TIME_IMPLEMENTATION(static, 1);
  TIME_IMPLEMENTATION(static, 10);
  TIME_IMPLEMENTATION(static, 100);
  TIME_IMPLEMENTATION(static, 1000);
  TIME_IMPLEMENTATION(static, 10000);
  TIME_IMPLEMENTATION(static, 100000);
  TIME_IMPLEMENTATION(static, 1000000);
  TIME_IMPLEMENTATION(static, 10000000);

  TIME_IMPLEMENTATION(dynamic, 1);
  TIME_IMPLEMENTATION(dynamic, 10);
  TIME_IMPLEMENTATION(dynamic, 100);
  TIME_IMPLEMENTATION(dynamic, 1000);
  TIME_IMPLEMENTATION(dynamic, 10000);
  TIME_IMPLEMENTATION(dynamic, 100000);
  TIME_IMPLEMENTATION(dynamic, 1000000);
  TIME_IMPLEMENTATION(dynamic, 10000000);

  TIME_IMPLEMENTATION(guided, 1);
  TIME_IMPLEMENTATION(guided, 10);
  TIME_IMPLEMENTATION(guided, 100);
  TIME_IMPLEMENTATION(guided, 1000);
  TIME_IMPLEMENTATION(guided, 10000);
  TIME_IMPLEMENTATION(guided, 100000);
  TIME_IMPLEMENTATION(guided, 1000000);
  TIME_IMPLEMENTATION(guided, 10000000);

#undef TIME_IMPLEMENTATION
#undef DO_PRAGMA

  free(u);
  free(v);

  exit(EXIT_SUCCESS);
}
